import os
import random
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def getRandomColor():
    """获取一个随机的颜色"""
    colorArr = ['1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F']
    color = ""
    for i in range(6):
        color += colorArr[random.randint(0, 14)]
    return "#" + color


class NodeItem:
    def __init__(self, nodeName, nodeId):
        self.nodeName = nodeName
        self.nodeId = nodeId


class GraphBase:
    """
    一个基于 matplotlib 实现的通用绘图接口
    """

    def __init__(self):
        # 记录所有可用的 Marker
        # https://matplotlib.org/stable/api/markers_api.html
        self.markers = ["o", "v", "^", "<", ">", "1", "2", "3", "4", "8", "s", "p", "P", "*", "h", "H", "+",
                        "x", "X", "D", "d", "|", "_", ".", ","]
        # 记录当前图已经使用的Marker
        self.currentUsedMarker = {}
        # 记录常用的 color
        self.colors = [
            "tab:blue", "tab:orange", "tab:green", "tab:red", "tab:purple",
            "tab:brown", "tab:pink", "tab:gray", "tab:olive", "tab:cyan", "red", "green", "blue", "c", "m", "y",
        ]
        # 记录当前图已经使用的 color
        self.currentUsedColor = {}
        # 记录常用的 hatch
        self.hatches = ['', '/', 'o', '\\', 'O', '|', '-', '+', 'x', '.', '*',
                        '/o', '\\|', '|*', '-\\', '+o', 'x*', 'o-', 'O|', 'O.', '*-']
        self.currentUsedHatch = {}

    def autoSelectHatch(self) -> str:
        """
        从所有的候选 Hatch 中自动选择
        :return:
        """
        for hatch in self.hatches:
            if hatch not in self.currentUsedHatch:
                self.currentUsedHatch[hatch] = hatch
                return hatch
        return ""

    def autoSelectMarker(self) -> str:
        """
        从所有的marker中自动选择 Marker，保证不和已有的Marker重复
        :return:
        """
        for marker in self.markers:
            if marker not in self.currentUsedMarker:
                self.currentUsedMarker[marker] = marker
                return marker
        return ""

    def autoSelectColor(self) -> str:
        for color in self.colors:
            if color not in self.currentUsedColor:
                return color
        while True:
            randomColor = getRandomColor()
            if randomColor not in self.currentUsedColor:
                break
            self.currentUsedColor[randomColor] = randomColor
        return randomColor

    def xlim(self, *args, **kwargs):
        plt.xlim(*args, **kwargs)
        return self

    def ylim(self, *args, **kwargs):
        plt.ylim(*args, **kwargs)
        return self

    def xticks(self, ticks=None, labels=None, **kwargs):
        plt.xticks(ticks, labels, **kwargs)
        return self

    def title(self, label, fontdict=None, loc=None, pad=None, **kwargs):
        plt.title(label, fontdict=fontdict, loc=loc, pad=pad, **kwargs)
        return self

    def xlabel(self, xlabel, fontdict=None, labelpad=None, loc=None, **kwargs):
        plt.xlabel(xlabel, fontdict, labelpad, loc=loc, **kwargs)
        return self

    def ylabel(self, ylabel, fontdict=None, labelpad=None, loc=None, **kwargs):
        plt.ylabel(ylabel, fontdict, labelpad, loc=loc, **kwargs)
        return self

    def legend(self, *args, **kwargs):
        plt.legend(*args, **kwargs)
        return self

    def innerPlot(self, x, y, *args,
                  color: str = "&&",
                  linewidth: float = 2, linestyle: str = "dotted", marker: str = "&&",
                  markerfacecolor: str = "none", markersize: float = 6,
                  **kwargs):
        if "linewidth" not in kwargs:
            kwargs["linewidth"] = linewidth
        if "linestyle" not in kwargs:
            kwargs["linestyle"] = linestyle
        # 默认自动选择Marker
        if marker == "&&":
            marker = self.autoSelectMarker()
            self.currentUsedMarker[marker] = marker
        if "marker" not in kwargs:
            kwargs["marker"] = marker
        if "markerfacecolor" not in kwargs:
            kwargs["markerfacecolor"] = markerfacecolor
        if "markersize" not in kwargs:
            kwargs["markersize"] = markersize
        # 默认自动选择 color
        if color == "&&":
            color = self.autoSelectColor()
            self.currentUsedColor[color] = color
        if "color" not in kwargs:
            kwargs["color"] = color
        plt.plot(x, y, *args, **kwargs)
        return self

    def innerPlotSum(self, nodeList: [NodeItem], getNode, getX, getY, *args,
                     color: str = "&&",
                     linewidth: float = 2, linestyle: str = "dotted", marker: str = "&&",
                     markerfacecolor: str = "none", markersize: float = 6,
                     **kwargs):
        xMerge = []
        yArr = []
        for nodeItem in nodeList:
            node = getNode(nodeItem.nodeName)
            if not node:
                print("not exist node: ", node)
                continue

            x, y = getX(node, nodeItem.nodeId), \
                   getY(node, nodeItem.nodeId)

            if len(x) > len(xMerge):
                xMerge = x
            yArr.append(np.array(y))

        if len(yArr) <= 0:
            return self

        yMerge = np.zeros(len(xMerge))
        # pad y
        for y in yArr:
            y = np.pad(y, (0, len(xMerge) - len(y)), 'constant', constant_values=(0, 0))
            yMerge += y

        if "label" not in kwargs:
            kwargs["label"] = "Merge"

        self.innerPlot(xMerge, yMerge, *args,
                       color=color,
                       linewidth=linewidth,
                       linestyle=linestyle,
                       marker=marker,
                       markerfacecolor=markerfacecolor,
                       markersize=markersize,
                       **kwargs)
        return self

    def innerPlotAvg(self, nodeList: [NodeItem], getNode, getX, getY, *args,
                     color: str = "&&",
                     linewidth: float = 2, linestyle: str = "dotted", marker: str = "&&",
                     markerfacecolor: str = "none", markersize: float = 6,
                     **kwargs):
        xMerge = []
        yArr = []

        for nodeItem in nodeList:
            node = getNode(nodeItem.nodeName)
            x, y = getX(node, nodeItem.nodeId), \
                   getY(node, nodeItem.nodeId)
            if not node:
                print("not exist node: ", node)
                continue
            if len(x) > len(xMerge):
                xMerge = x
            yArr.append(np.array(y))

        if len(yArr) <= 0:
            return self

        yMerge = np.zeros(len(xMerge))

        # pad y
        for y in yArr:
            y = np.pad(y, (0, len(xMerge) - len(y)), 'constant', constant_values=(0, 0))
            yMerge += y
        yMerge /= len(yArr)

        if "label" not in kwargs:
            kwargs["label"] = "Merge"

        self.innerPlot(xMerge, yMerge, *args,
                       color=color,
                       linewidth=linewidth,
                       linestyle=linestyle,
                       marker=marker,
                       markerfacecolor=markerfacecolor,
                       markersize=markersize,
                       **kwargs)
        return self

    def show(self, *args, **kwargs):
        """
        绘制吞吐量图，并show
        :return:
        """
        plt.show(*args, **kwargs)
        return self

    def drawAndSave(self, dir: str, name: str, *args, **kwargs):
        """
        绘制吞吐量图，并保存到图片当中
        :return:
        """
        if not os.path.exists(dir):
            os.makedirs(dir)
        plt.savefig(f"{dir}{os.sep}{name}", *args, **kwargs)
        return self

    def close(self):
        plt.close()


class ExportItem:
    def __init__(self, nodeName: str,
                 units: str, label: str, x: [float], y: [float]):
        self.nodeName = nodeName
        self.units = units
        self.label = label
        self.x = x
        self.y = y


class ExportBase:
    """
    导出基类，用于将统计数据导出到Excel当中
    """

    def __init__(self, getNode, getX, getY, getUnit):
        self.items: [ExportItem] = []
        self.getNode = getNode
        self.getX = getX
        self.getY = getY
        self.getUnit = getUnit

    def select(self, nodeName: str, faceId: any, label: str = None):
        node = self.getNode(nodeName)
        if not node:
            print("not exist node: ", node)
            return self
        if label is None:
            label = nodeName
        x, y = self.getX(node, faceId), self.getY(node, faceId)
        self.items.append(
            ExportItem(
                nodeName,
                self.getUnit(),
                label,
                x,
                y))
        return self

    def selectAvg(self, nodeList: [NodeItem], label: str = None):
        xMerge = []
        yArr = []

        for nodeItem in nodeList:
            node = self.getNode(nodeItem.nodeName)
            x, y = self.getX(node, nodeItem.nodeId), self.getY(node, nodeItem.nodeId)
            if not node:
                print("not exist node: ", node)
                continue
            if len(x) > len(xMerge):
                xMerge = x
            yArr.append(np.array(y))

        if len(yArr) <= 0:
            return self

        yMerge = np.zeros(len(xMerge))

        # pad y
        for y in yArr:
            y = np.pad(y, (0, len(xMerge) - len(y)), 'constant', constant_values=(0, 0))
            yMerge += y
        yMerge /= len(yArr)
        self.items.append(
            ExportItem(
                label,
                self.getUnit(),
                label,
                xMerge,
                yMerge))
        return self

    def selectSum(self, nodeList: [NodeItem], label: str = None):
        xMerge = []
        yArr = []

        for nodeItem in nodeList:
            node = self.getNode(nodeItem.nodeName)
            x, y = self.getX(node, nodeItem.nodeId), self.getY(node, nodeItem.nodeId)
            if not node:
                print("not exist node: ", node)
                continue
            if len(x) > len(xMerge):
                xMerge = x
            yArr.append(np.array(y))

        if len(yArr) <= 0:
            return self

        yMerge = np.zeros(len(xMerge))

        # pad y
        for y in yArr:
            y = np.pad(y, (0, len(xMerge) - len(y)), 'constant', constant_values=(0, 0))
            yMerge += y
        self.items.append(
            ExportItem(
                label,
                self.getUnit(),
                label,
                xMerge,
                yMerge))
        return self

    def exportCDFExcel(self, savePath: str,
                       xName: str = "Rate",
                       yName: str = "CDF"):
        """
        计算CDF（Cumulative Distribution Function）并导出
        :param xName:
        :param yName:
        :param savePath:
        :return:
        """
        longNames, units, labels = [], [], []
        ys, maxLen = [], 0

        for item in self.items:
            maxLen = max(maxLen, len(item.y))

        for item in self.items:
            y = np.array(np.pad(item.y, (0, maxLen - len(item.y)), 'constant', constant_values=(0, 0))).T
            x = np.sort(y)
            ys.append(x)
            longNames.append(xName)
            units.append(item.units)
            labels.append(item.label)
            y = 1. * np.arange(len(x)) / (len(x) - 1)
            ys.append(y)
            longNames.append(yName)
            units.append("")
            labels.append(item.label)


        contents = []
        # long name
        contents.append(longNames)

        # units
        contents.append(units)

        # comments
        contents.append(labels)

        dataMatrix = np.vstack(ys)

        # write datas
        for i in range(maxLen):
            contents.append(dataMatrix[:, i])

        df = pd.DataFrame(contents)
        df.to_excel(savePath, index=False, header=None)
        return self

    def exportExcel(self, savePath: str,
                    xName: str = "Times",
                    yName: str = "Rate",
                    xUnit: str = "Second"):
        """
        导出数据到Excel
        :param savePath:
        :param xName:
        :param yName:
        :param xUnit:
        :return:
        """
        contents = []
        nodeNames, units, labels, xResult = [], [], [], []
        for item in self.items:
            nodeNames.append(item.nodeName)
            units.append(item.units)
            labels.append(item.label)
            if len(item.x) > len(xResult):
                xResult = item.x
        # long name
        contents.append([xName] + [yName for _ in range(len(nodeNames))])

        # units
        contents.append([xUnit] + units)

        # comments
        contents.append([""] + labels)

        ys = [xResult]
        # padding y
        for item in self.items:
            ys.append(np.array(np.pad(item.y, (0, len(xResult) - len(item.y)), 'constant', constant_values=(0, 0))).T)

        dataMatrix = np.vstack(ys)

        # write datas
        for i in range(len(xResult)):
            contents.append(dataMatrix[:, i])

        df = pd.DataFrame(contents)
        df.to_excel(savePath, index=False, header=None)
        return self



from enum import Enum
from typing import Optional
from .common import GraphBase, NodeItem, ExportItem, ExportBase
import numpy as np
import pandas as pd


class ThroughputType(Enum):
    """
    吞吐量类型
    https://ndnsim.net/current/metric.html
    """
    InInterests = 0
    OutInterests = 1
    InData = 2
    OutData = 3
    InNacks = 4
    OutNacks = 5
    InSatisfiedInterests = 6
    InTimedOutInterests = 7
    OutSatisfiedInterests = 8
    OutTimedOutInterests = 9

    # 下面的类型只在 FaceId = -1 有效
    SatisfiedInterests = 10
    TimedOutInterests = 11


class ThroughputTarget(Enum):
    """
    吞吐量目标
    https://ndnsim.net/current/metric.html
    """
    # estimated rate (EWMA average) of packets within the last averaging period (number of packets/s)
    Packets = 0

    # estimated rate (EWMA average) within last averaging period (kilobytes/s)
    Kilobytes_KBps = 1
    Kilobytes_MBps = 2
    Kilobytes_Kbps = 3
    Kilobytes_Mbps = 4

    # absolute number of packets within last averaging period (number of packets).
    PacketRaw = 5

    # absolute number of kilobytes transferred within the last averaging period (number of packets).
    KilobytesRaw_KBps = 6
    KilobytesRaw_MBps = 7
    KilobytesRaw_Kbps = 8
    KilobytesRaw_Mbps = 9

    @staticmethod
    def getUnit(throughputTargetName: str) -> str:
        value = ThroughputTarget[throughputTargetName].value
        if value == 0 or value == 5:
            return "Number of packets"
        elif value == 1 or value == 6:
            return "KBps"
        elif value == 2 or value == 7:
            return "MBps"
        elif value == 3 or value == 8:
            return "Kbps"
        elif value == 4 or value == 9:
            return "Mbps"
        else:
            return "unknown"


class ThroughputItem:
    """
    一个用于描述 ndnsim 吞吐量采样结果的类
    """

    def __init__(self, Time: float, Node: str, FaceId: int, FaceDescr: str, Type: ThroughputType,
                 Packets: float, Kilobytes: float, PacketRaw: float, KilobytesRaw: float):
        """
        :param Time:        时间，例如: 0.5
        :param Node:        节点名称，例如：C1
        :param FaceId:      Face id
        :param FaceDescr:   Face 描述
        :param Type:
        :param Packets:
        :param Kilobytes:
        :param PacketRaw:
        :param KilobytesRaw:
        """
        self.Time = Time  # simulation time
        self.Node = Node  # node id, globally unique
        self.FaceId = FaceId  # interface ID (-1 for combined metric)
        self.FaceDescr = FaceDescr
        self.Type = Type
        self.Packets = Packets
        self.Kilobytes = Kilobytes
        self.PacketRaw = PacketRaw
        self.KilobytesRaw = KilobytesRaw

    def getValueByThroughputTarget(self, throughputTarget: ThroughputTarget, lpPacketHeaderSize: int = 0):
        kilobytes, kilobytesRaw = self.Kilobytes, self.KilobytesRaw
        if kilobytes > 0:
            kilobytes += ((self.Packets * lpPacketHeaderSize) / 1000)
        if kilobytesRaw > 0:
            kilobytesRaw += ((self.PacketRaw * lpPacketHeaderSize) / 1000)
        if throughputTarget == ThroughputTarget.Packets:
            return self.Packets
        elif throughputTarget == ThroughputTarget.Kilobytes_KBps:
            return kilobytes
        elif throughputTarget == ThroughputTarget.Kilobytes_MBps:
            return kilobytes / 1000
        elif throughputTarget == ThroughputTarget.Kilobytes_Kbps:
            return kilobytes * 8
        elif throughputTarget == ThroughputTarget.Kilobytes_Mbps:
            return kilobytes * 8 / 1000
        elif throughputTarget == ThroughputTarget.PacketRaw:
            return self.PacketRaw
        elif throughputTarget == ThroughputTarget.KilobytesRaw_KBps:
            return kilobytesRaw
        elif throughputTarget == ThroughputTarget.KilobytesRaw_MBps:
            return kilobytesRaw / 1000
        elif throughputTarget == ThroughputTarget.KilobytesRaw_Kbps:
            return kilobytesRaw * 8
        elif throughputTarget == ThroughputTarget.KilobytesRaw_Mbps:
            return kilobytesRaw * 8 / 1000

    @staticmethod
    def parseLine(line: str):
        # 0.5	C1	1	internal://	InInterests	0	0	0	0
        values = line.strip().split("\t")
        if len(values) != 9:
            return None
        return ThroughputItem(float(values[0].strip()),
                              values[1].strip(),
                              int(values[2].strip()),
                              values[3].strip(),
                              ThroughputType[values[4].strip()],
                              float(values[5].strip()),
                              float(values[6].strip()),
                              float(values[7].strip()),
                              float(values[8].strip()),
                              )


class NodeFaceThroughput:
    """
    一个用于描述节点某个接口吞吐量的类
    """

    def __init__(self, Node: str, FaceId: int):
        self.Node = Node
        self.FaceId = FaceId
        self.typeMap = dict()  # 用于存储不同 Throughput Type 的采样记录
        for throughputType in ThroughputType:
            self.typeMap[throughputType.name] = []

    def appendItem(self, item: ThroughputItem):
        """
        新增一条采样记录
        :param item:
        :return:
        """
        # 如果不是当前接口的统计记录，忽略
        if item.Node != self.Node or item.FaceId != self.FaceId:
            return
        self.typeMap[item.Type.name].append(item)

    def getX(self, throughputType: ThroughputType = ThroughputType.OutData, samplingInterval: float = 1.0):
        """
        获取采样时间列表
        :param throughputType:
        :param samplingInterval:
        :return:
        """
        lastCount, res = 0, []
        if len(self.typeMap[throughputType.name]) <= 0:
            return []
        item = self.typeMap[throughputType.name][0]
        if item.Time / samplingInterval > 0:
            for i in range(int(item.Time / samplingInterval)):
                res.append((i + 1) * samplingInterval)
                lastCount = i + 1
        for item in self.typeMap[throughputType.name]:
            if int(item.Time / samplingInterval) != lastCount:
                lastCount = int(item.Time / samplingInterval)
            else:
                continue
            res.append(lastCount * samplingInterval)
        return res

    def getY(self, throughputType: ThroughputType = ThroughputType.OutData,
             throughputTarget: ThroughputTarget = ThroughputTarget.Kilobytes_KBps,
             samplingInterval: float = 1.0, lpPacketHeaderSize: int = 0):
        """
        获取吞吐量列表
        :param lpPacketHeaderSize:      LpPacket Header 大小
        :param throughputType:          吞吐量类型
        :param throughputTarget:        吞吐量目标
        :param samplingInterval:       采样间隔
        :return:
        """
        lastCount, res = 0, []
        if len(self.typeMap[throughputType.name]) <= 0:
            return []
        item = self.typeMap[throughputType.name][0]
        if item.Time / samplingInterval > 0:
            for i in range(int(item.Time / samplingInterval)):
                res.append(0)
                lastCount = i + 1
        for item in self.typeMap[throughputType.name]:
            if int(item.Time / samplingInterval) != lastCount:
                lastCount = int(item.Time / samplingInterval)
            else:
                continue
            res.append(item.getValueByThroughputTarget(throughputTarget, lpPacketHeaderSize))
        return res


class NodeThroughput:
    """
    一个用于描述节点各个接口吞吐量的类
    """

    def __init__(self, Node: str):
        self.faceIdMap = dict()
        self.Node = Node

    def appendItem(self, item: ThroughputItem):
        """
        新增一条采样记录
        :param item:
        :return:
        """
        # 如果不是当前节点的统计记录，忽略
        if item.Node != self.Node:
            return
        if item.FaceId not in self.faceIdMap:
            self.faceIdMap[item.FaceId] = NodeFaceThroughput(item.Node, item.FaceId)
        self.faceIdMap[item.FaceId].appendItem(item)

    def getX(self, FaceId: int, throughputType: ThroughputType = ThroughputType.OutData,
             samplingInterval: float = 1.0) -> [float]:
        """
        获取采样时间列表
        :param FaceId:
        :param throughputType:
        :param samplingInterval:
        :return:
        """
        if FaceId not in self.faceIdMap:
            return []
        return self.faceIdMap[FaceId].getX(throughputType, samplingInterval)

    def getY(self, FaceId: int, throughputType: ThroughputType = ThroughputType.OutData,
             throughputTarget: ThroughputTarget = ThroughputTarget.Kilobytes_KBps, samplingInterval: float = 1.0,
             lpPacketHeaderSize: int = 0):
        """
        获取吞吐量列表
        :param lpPacketHeaderSize:      LpPacket Header 大小
        :param FaceId:                  FaceId
        :param throughputType:          吞吐量类型
        :param throughputTarget:        吞吐量目标
        :param samplingInterval:       采样间隔
        :return:
        """
        if FaceId not in self.faceIdMap:
            return []
        return self.faceIdMap[FaceId].getY(throughputType, throughputTarget, samplingInterval, lpPacketHeaderSize)


class Throughput:
    """
    一个用于解析 ndnsim 吞吐量采样结果的类
    """

    def __init__(self):
        self.nodeMap = dict()

    def getByNode(self, node: str) -> Optional[NodeThroughput]:
        if node not in self.nodeMap:
            return None
        return self.nodeMap[node]

    @staticmethod
    def parse(inputFile: str):
        """
        传入一个 ndnsim 采集的吞吐量结果文件，从中解析出吞吐量
        :param inputFile:
        :return:
        """
        throughput = Throughput()
        with open(inputFile, "r") as f:
            # 首先忽略表头
            # Time	Node	FaceId	FaceDescr	Type	Packets	Kilobytes	PacketRaw	KilobytesRaw
            f.readline()

            # 从第一行开始解析
            line = f.readline()
            while line:
                if line.strip() != "":
                    item = ThroughputItem.parseLine(line)
                    if item:
                        if item.Node not in throughput.nodeMap:
                            throughput.nodeMap[item.Node] = NodeThroughput(item.Node)
                        throughput.nodeMap[item.Node].appendItem(item)
                line = f.readline()
        return throughput


class ThroughputExtractBase:
    def __init__(self, throughput: Throughput):
        self.samplingInterval = 1.0
        self.throughputType = ThroughputType.OutData
        self.throughputTarget = ThroughputTarget.Kilobytes_KBps
        self.lpPacketHeaderSize = 0  # LpPacket 头部大小
        self.throughput = throughput

    def setSamplingInterval(self, samplingInterval: float):
        """
        设置采样间隔 => 如果 samplingInterval = 1.0 表示绘图时，每秒采样一次
        :param samplingInterval:
        :return:
        """
        self.samplingInterval = samplingInterval
        return self

    def setThroughputType(self, throughputType: ThroughputType):
        """
        设置吞吐量类型
        :param throughputType:
        :return:
        """
        self.throughputType = throughputType
        return self

    def setThroughputTarget(self, throughputTarget: ThroughputTarget):
        """
        设置吞吐量目标
        :param throughputTarget:
        :return:
        """
        self.throughputTarget = throughputTarget
        return self

    def setLpPacketHeaderSize(self, lpPacketHeaderSize: int):
        """
        设置LpPacket头部大小
        :param lpPacketHeaderSize:
        :return:
        """
        self.lpPacketHeaderSize = lpPacketHeaderSize
        return self

    def getNode(self, nodeName):
        return self.throughput.getByNode(nodeName)

    def getX(self, node: NodeThroughput, nodeId):
        return node.getX(nodeId, samplingInterval=self.samplingInterval)

    def getY(self, node: NodeThroughput, nodeId):
        return node.getY(nodeId, self.throughputType, self.throughputTarget, self.samplingInterval,
                         self.lpPacketHeaderSize)

    def getUnit(self):
        return ThroughputTarget.getUnit(self.throughputTarget.name)


class ThroughputExport(ThroughputExtractBase, ExportBase):
    """
    一个用于导出数据到Excel的辅助类
    """

    def __init__(self, throughput: Throughput):
        ThroughputExtractBase.__init__(self, throughput)
        ExportBase.__init__(self, self.getNode, self.getX, self.getY, self.getUnit)

    def innerParse(self, inputFile: str):
        self.throughput = Throughput.parse(inputFile)
        return self

    @staticmethod
    def parse(inputFile: str):
        return ThroughputExport(Throughput.parse(inputFile))


class ThroughputGraph(GraphBase, ThroughputExtractBase):
    """
    一个用于实现绘制吞吐量图的类
    """

    def __init__(self, throughput: Throughput):
        GraphBase.__init__(self)
        ThroughputExtractBase.__init__(self, throughput)

    def plot(self, node: str, faceId: int, *args,
             color: str = "&&",
             linewidth: float = 2, linestyle: str = "dotted", marker: str = "&&",
             markerfacecolor: str = "none", markersize: float = 6,
             **kwargs):
        node = self.throughput.getByNode(node)
        if not node:
            print("not exist node: ", node)
            return self
        if "label" not in kwargs:
            kwargs["label"] = node.Node
        x, y = node.getX(faceId, samplingInterval=self.samplingInterval), \
               node.getY(faceId, self.throughputType, self.throughputTarget, self.samplingInterval,
                         self.lpPacketHeaderSize)
        super().innerPlot(x, y, *args,
                          color=color,
                          linewidth=linewidth,
                          linestyle=linestyle,
                          marker=marker,
                          markerfacecolor=markerfacecolor,
                          markersize=markersize,
                          **kwargs)
        return self

    def plotSum(self, nodeList: [NodeItem], *args,
                color: str = "&&",
                linewidth: float = 2, linestyle: str = "dotted", marker: str = "&&",
                markerfacecolor: str = "none", markersize: float = 6,
                **kwargs):

        super().innerPlotSum(nodeList, self.getNode, self.getX, self.getY, *args,
                             color=color,
                             linewidth=linewidth,
                             linestyle=linestyle,
                             marker=marker,
                             markerfacecolor=markerfacecolor,
                             markersize=markersize,
                             **kwargs)
        return self

    def plotAvg(self, nodeList: [NodeItem], *args,
                color: str = "&&",
                linewidth: float = 2, linestyle: str = "dotted", marker: str = "&&",
                markerfacecolor: str = "none", markersize: float = 6,
                **kwargs):

        self.innerPlotAvg(nodeList, self.getNode, self.getX, self.getY, *args,
                          color=color,
                          linewidth=linewidth,
                          linestyle=linestyle,
                          marker=marker,
                          markerfacecolor=markerfacecolor,
                          markersize=markersize,
                          **kwargs)
        return self

    def innerParse(self, inputFile: str):
        self.throughput = Throughput.parse(inputFile)
        return self

    @staticmethod
    def parse(inputFile: str):
        return ThroughputGraph(Throughput.parse(inputFile))



from .common import GraphBase, getRandomColor
from .throughput import Throughput, NodeItem, ThroughputType, ThroughputTarget
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.legend import Legend
from matplotlib.patches import Patch
import matplotlib.ticker as mticker


class DeliveryTimeCollection:
    """
    分发时间集合，包含了期望从同一次实验获取的所有内容交付时间数据
    """

    def __init__(self, collectionName: str, labels: [str], colors: [str],
                 contentDeliveryTimeGroups: [[float]], hatch: str,
                 xlabels: [str]):
        """
        构造一个分发时间集合
        :param collectionName:      集合的名字，通常是方案的名字，比如 DSCCP
        :param labels:              标签列表
        :param contentDeliveryTimeGroups:   分发节点条目
        """
        self.collectionName = collectionName
        self.labels = labels
        self.colors = colors
        self.xlabels = xlabels
        self.hatch = hatch
        self.contentDeliveryTimeGroups = contentDeliveryTimeGroups


class DeliveryTimeCollectionBuilder:
    def __init__(self, collectionName: str, throughputFile: str, hatch: str = None):
        self.collectionName = collectionName
        self.throughput = Throughput.parse(throughputFile)
        self.labels = []
        self.deliveryTimeGroups = []
        self.hatch = hatch
        self.xlabels = []
        self.usedColor = []
        # 记录常用的 color
        self.colors = [
            "tab:blue", "tab:orange", "tab:green", "tab:red", "tab:purple",
            "tab:brown", "tab:pink", "tab:gray", "tab:olive", "tab:cyan", "red", "green", "blue", "c", "m", "y",
        ]
        # 记录当前图已经使用的 color
        self.currentUsedColor = {}

    def autoSelectColor(self) -> str:
        for color in self.colors:
            if color not in self.currentUsedColor:
                self.currentUsedColor[color] = color
                return color
        while True:
            randomColor = getRandomColor()
            if randomColor not in self.currentUsedColor:
                break
            self.currentUsedColor[randomColor] = randomColor
        return randomColor

    def appendDeliveryNodeItem(self, label: str, items: [NodeItem], color: str = "&&", samplingInterval=0.5,
                               throughputType: ThroughputType = ThroughputType.OutData,
                               throughputTarget: ThroughputTarget = ThroughputTarget.PacketRaw):
        self.labels.append(label)
        if color == "&&":
            color = self.autoSelectColor()
        self.usedColor.append(color)
        contentDeliveryTimes, xlabels = [], []
        for item in items:
            contentDeliveryTime = 0
            x = self.throughput.getByNode(item.nodeName).getX(item.nodeId,
                                                              samplingInterval=samplingInterval,
                                                              throughputType=throughputType)
            y = self.throughput.getByNode(item.nodeName).getY(item.nodeId,
                                                              samplingInterval=samplingInterval,
                                                              throughputType=throughputType,
                                                              throughputTarget=throughputTarget)
            if len(x) > 0:
                contentDeliveryTime = x[-1]
                for i in reversed(range(len(y))):
                    if y[i] != 0:
                        contentDeliveryTime = x[i]
                        break
            contentDeliveryTimes.append(contentDeliveryTime)
            xlabels.append(item.nodeName)
        self.xlabels.append(xlabels)
        self.deliveryTimeGroups.append(contentDeliveryTimes)
        return self

    def build(self) -> DeliveryTimeCollection:
        return DeliveryTimeCollection(self.collectionName, self.labels, self.usedColor,
                                      self.deliveryTimeGroups, self.hatch, self.xlabels)


class DeliveryTimeGraph(GraphBase):
    """
    一个用于绘制内容交付时间图的类
    1. 通过吞吐量统计数据进行判断 => 如果一个 Consumer 的吞吐量变成0了，认为内容交付完成，取该时间为内容交付时间
    """

    def __init__(self):
        super(DeliveryTimeGraph, self).__init__()
        self.deliveryTimeCollections: [DeliveryTimeCollection] = []
        self.figSize = (10, 7)  # 图的大小
        self.edgecolor = "black"  # 边缘颜色
        self.totalWidth = 1
        self.fig = None
        self.ax = None

    def setTotalWidth(self, totalWidth: float):
        self.totalWidth = totalWidth
        return self

    def appendDeliveryTimeCollection(self, deliveryTimeCollection: DeliveryTimeCollection):
        """
        添加一个内容完成时间条目
        :return:
        """
        if deliveryTimeCollection.hatch is None:
            deliveryTimeCollection.hatch = self.autoSelectHatch()
        self.deliveryTimeCollections.append(deliveryTimeCollection)
        return self

    def xlim(self, *args, **kwargs):
        self.ax.set_xlim(*args, **kwargs)
        return self

    def ylim(self, *args, **kwargs):
        plt.ylim(*args, **kwargs)
        return self

    def xticks(self, ticks=None, labels=None, **kwargs):
        plt.xticks(ticks, labels, **kwargs)
        return self

    def title(self, label, fontdict=None, loc=None, pad=None, **kwargs):
        self.ax.title(label, fontdict=fontdict, loc=loc, pad=pad, **kwargs)
        return self

    def legend(self, *args, **kwargs):
        handles, labels = self.ax.get_legend_handles_labels()
        self.ax.legend(handles, labels, *args, **kwargs)
        return self

    def categoryLegend(self, facecolor='white', *args, loc='upper left', **kwargs):
        categories, patches = [], []
        for dtc in self.deliveryTimeCollections:
            categories.append(dtc.collectionName)
            patches.append(Patch(facecolor=facecolor, edgecolor=self.edgecolor, hatch=dtc.hatch))
        if "frameon" not in kwargs:
            kwargs["frameon"] = False
        leg = Legend(self.ax, patches, categories, *args, loc=loc, **kwargs)
        self.ax.add_artist(leg)
        return self

    def doDraw(self):
        """
        绘制内容完成时间图
        :return:
        """
        maxSize, maxN = 0, 0
        for dtc in self.deliveryTimeCollections:
            if len(dtc.contentDeliveryTimeGroups[0]) > maxSize:
                maxSize = len(dtc.contentDeliveryTimeGroups[0])
            if len(dtc.contentDeliveryTimeGroups) > maxN:
                maxN = len(dtc.contentDeliveryTimeGroups)
        x = np.arange(maxSize)
        total_width, n = self.totalWidth, maxN
        width = total_width / n
        x = x - ((len(self.deliveryTimeCollections) - 1) * width / 2)

        self.fig, self.ax = plt.subplots(figsize=self.figSize)
        offsetX, isFirst = 0, True
        for dtc in self.deliveryTimeCollections:
            bottom = np.zeros(len(dtc.contentDeliveryTimeGroups[0]))
            for i in range(len(dtc.colors)):
                if isFirst:
                    self.ax.bar(x + offsetX, dtc.contentDeliveryTimeGroups[i], width=width, color=dtc.colors[i],
                                edgecolor=self.edgecolor,
                                label=dtc.labels[i],
                                bottom=bottom,
                                hatch=dtc.hatch)
                else:
                    self.ax.bar(x + offsetX, dtc.contentDeliveryTimeGroups[i], width=width, color=dtc.colors[i],
                                edgecolor=self.edgecolor,
                                bottom=bottom,
                                hatch=dtc.hatch)
                bottom += np.array(dtc.contentDeliveryTimeGroups[i])
            isFirst = False
            offsetX += width

        consumers = self.deliveryTimeCollections[0].xlabels[0]
        ticks_loc = self.ax.get_xticks().tolist()
        self.ax.xaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
        self.ax.set_xticklabels([""] + consumers + [""])
        return self
